/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <guichaz@yahoo.fr>
 */

/************************************
 * The Images and Directories menus *
 ************************************/

#include <string.h>             /* strlen() */

#include "gliv.h"
#include "images_menus.h"
#include "options.h"
#include "menus.h"
#include "messages.h"
#include "next_image.h"
#include "mnemonics.h"
#include "files_list.h"
#include "loading.h"
#include "thumbnails.h"
#include "callbacks.h"
#include "tree.h"
#include "timestamp.h"

extern options_struct *options;

/* The progress indicator. */
static GtkMenuItem *rebuilding_entry;

/* The menu entry to cancel the rebuilding. */
static GtkMenuItem *cancel_menu_item;

/*** Functions for both menus. ***/

/* Called at menu creation time. */
void set_rebuilding_entry(GtkMenuItem * item)
{
    rebuilding_entry = item;
}

/* Refresh the percent indicator. */
static void set_menu_indicator(gchar * name, gint percent)
{
    static gchar *rebuilding = NULL;

    if (name == NULL)
        set_menu_label(rebuilding_entry, "", FALSE);
    else {
        gchar *label;

        if (rebuilding == NULL)
            /* First time. */
            rebuilding = _("Rebuilding:");

        label = g_strdup_printf("%s %s (%d%%)", rebuilding, name, percent);
        set_menu_label(rebuilding_entry, label, FALSE);

        g_free(label);
    }
}

/* Update the percent indicator. */
static void set_progress(gchar * menu, gint * percent, gint number)
{
    gint length;

    length = get_list_length();

    if ((*percent + 1) * length <= 100 * number) {
        *percent = (100 * number) / length;
        set_menu_indicator(menu, *percent);
    }
}

/* We don't want to perform the signal lookup for each file. */
static void connect_activate(GtkMenuItem * instance, const gchar * filename)
{
    static guint signal_id = 0;
    GClosure *closure;

    if (signal_id == 0)
        /* First time. */
        signal_id = g_signal_lookup("activate", G_TYPE_FROM_INSTANCE(instance));

    closure = g_cclosure_new_swap(G_CALLBACK(menu_load),
                                  (gpointer) filename, NULL);

    g_signal_connect_closure_by_id(instance, signal_id, 0, closure, FALSE);
}

typedef enum {
    MENU_FILE,                  /* From both menus. */
    MENU_DIR,                   /* From the Images menu. */
    MENU_SUBMENU                /* From the Directories menu. */
} menu_type;

/* Add also a mnemonic and a thumbnail. */
static GtkMenuItem *new_menu_item(tree_item * item, menu_type type)
{
    const gchar *name;
    GtkMenuItem *menu_item;
    GtkImage *thumbnail = NULL;

    if (options->thumbnails && type == MENU_FILE && fill_thumbnail(item))
        thumbnail = GTK_IMAGE(gtk_image_new_from_pixbuf(item->thumb));
    else
        /* Simulate the do_threaded() effect of fill_thumbnail(). */
        process_events();

    if (type == MENU_DIR)
        name = item->path;
    else
        name = item->name;

    if (options->mnemonics && type != MENU_DIR) {
        name = add_mnemonic(name);

        if (thumbnail)
            menu_item =
                GTK_MENU_ITEM(gtk_image_menu_item_new_with_mnemonic(name));
        else
            menu_item = GTK_MENU_ITEM(gtk_menu_item_new_with_mnemonic(name));

    } else {
        if (thumbnail)
            menu_item = GTK_MENU_ITEM(gtk_image_menu_item_new_with_label(name));
        else
            menu_item = GTK_MENU_ITEM(gtk_menu_item_new_with_label(name));
    }

    if (thumbnail)
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menu_item),
                                      GTK_WIDGET(thumbnail));

    switch (type) {
    case MENU_FILE:
        connect_activate(menu_item, item->path);
        break;

    case MENU_DIR:
        gtk_widget_set_sensitive(GTK_WIDGET(menu_item), FALSE);
        break;

    case MENU_SUBMENU:
        break;
    }

    return menu_item;
}

/*
 * We first insert in the head of a list, and then we insert every elements
 * from that list in the menu head.  This way we build the menu in O(n) and
 * the mnemonics are created in normal order.
 */
static void insert_list_into_menu(GtkMenuShell * menu, GSList * list, gint pos)
{
    GSList *ptr;

    for (ptr = list; ptr != NULL; ptr = ptr->next)
        gtk_menu_shell_insert(menu, GTK_WIDGET(ptr->data), pos);

    g_slist_free(list);
}

/*** Directories menu. ***/

static void add_file_from_tree(GSList ** list, tree_item * item)
{
    static gchar *directories = NULL;
    static gint number = 0, percent = 0;

    if (canceled_using_tree())
        return;

    if (list == NULL) {
        /* First time for this menu. */

        if (directories == NULL)
            /* First time. */
            directories = _("Directories");

        number = 0;
        percent = 0;
        return;
    }

    *list = g_slist_prepend(*list, new_menu_item(item, MENU_FILE));

    set_progress(directories, &percent, number);
    number++;
}

static GtkMenuItem *add_sub_menu(GSList ** list, tree_item * item)
{
    GtkMenuShell *menu;
    GtkMenuItem *menu_item;

    menu_item = new_menu_item(item, MENU_SUBMENU);

    menu = GTK_MENU_SHELL(gtk_menu_new());
    gtk_menu_item_set_submenu(menu_item, GTK_WIDGET(menu));

    gtk_menu_shell_prepend(menu, gtk_tearoff_menu_item_new());
    *list = g_slist_prepend(*list, menu_item);

    return menu_item;
}

static void make_menu_from_tree_rec(GNode * tree, GSList ** list)
{
    GtkMenuItem *menu_item;
    tree_item *item;
    GSList *new_list = NULL;

    if (canceled_using_tree())
        return;

    item = tree->data;

    if (G_NODE_IS_LEAF(tree)) {
        add_file_from_tree(list, item);
        return;
    }

    menu_item = add_sub_menu(list, item);

    push_mnemonics();

    g_node_children_foreach(tree, G_TRAVERSE_ALL,
                            (GNodeForeachFunc) make_menu_from_tree_rec,
                            &new_list);

    pop_mnemonics();

    insert_list_into_menu(GTK_MENU_SHELL(gtk_menu_item_get_submenu(menu_item)),
                          new_list, 1);
}

/*** Images menu. ***/

static gboolean add_file_item(GNode * tree, GSList ** list)
{
    static gchar *images = NULL;
    static gint number = 0, percent = 0;
    GNode *sibling;

    if (canceled_using_tree())
        return TRUE;

    if (tree == NULL) {
        /* First time for this menu. */

        if (images == NULL)
            /* First time. */
            images = _("Images");

        number = 0;
        percent = 0;
        return FALSE;
    }


    /* Check if it is the first image in its directory. */
    sibling = g_node_prev_sibling(tree);
    if (!(G_NODE_IS_ROOT(tree) || (sibling && G_NODE_IS_LEAF(sibling))))
        *list = g_slist_prepend(*list,
                                new_menu_item(tree->parent->data, MENU_DIR));

    *list = g_slist_prepend(*list, new_menu_item(tree->data, MENU_FILE));

    set_progress(images, &percent, number);
    number++;

    return FALSE;
}

/*** Menus builders. ***/

static GtkWidget *common_rebuild(GtkMenuItem * root, gboolean start)
{
    static GtkMenuShell *menu;

    if (root != NULL)
        gtk_menu_item_deselect(root);

    if (start == FALSE) {
        const gchar *label;
        GtkMenuItem *item;

        label = add_mnemonic(_("Rebuild this menu"));
        item = GTK_MENU_ITEM(gtk_menu_item_new_with_mnemonic(label));

        gtk_menu_shell_insert(menu, GTK_WIDGET(item), 1);

        return GTK_WIDGET(item);
    }

    menu = GTK_MENU_SHELL(gtk_menu_new());
    gtk_menu_shell_prepend(menu, gtk_tearoff_menu_item_new());
    gtk_menu_item_set_submenu(root, GTK_WIDGET(menu));

    return GTK_WIDGET(menu);
}

static GNode *get_tree(void)
{
    if (is_loading())
        /*
         * There is no problem in rebuilding the images menus during
         * a loading, it would just make the loading too long.
         */
        return NULL;

    return make_tree();
}

#define REBUILD_START(name)                                                  \
    do {                                                                     \
        if (root != NULL) {                                                  \
            menu_item = root;                                                \
            return TRUE;                                                     \
        }                                                                    \
                                                                             \
        if (has_timestamp(timestamp) && last_tree_ok(timestamp))             \
            /* The menu is already up to date. */                            \
            return TRUE;                                                     \
                                                                             \
        tree = get_tree();                                                   \
        if (tree == NULL)                                                    \
            return FALSE;                                                    \
                                                                             \
        gtk_widget_set_sensitive(GTK_WIDGET(cancel_menu_item), TRUE);        \
        reset_mnemonics();                                                   \
        menu = GTK_MENU_SHELL(common_rebuild(menu_item, TRUE));              \
        set_menu_indicator(name, 0);                                         \
    } while (0)

#define REBUILD_END(function)                                                \
    do {                                                                     \
        GtkMenuItem *launcher;                                               \
                                                                             \
        launcher = GTK_MENU_ITEM(common_rebuild(NULL, FALSE));               \
        g_signal_connect_swapped(launcher, "activate", G_CALLBACK(function), \
                                 NULL);                                      \
                                                                             \
        insert_list_into_menu(menu, list, 2);                                \
        gtk_widget_show_all(GTK_WIDGET(menu_item));                          \
        set_menu_indicator(NULL, 0);                                         \
        gtk_widget_set_sensitive(GTK_WIDGET(cancel_menu_item), FALSE);       \
        if (canceled_using_tree())                                           \
            reset_timestamp(&timestamp);                                     \
        else                                                                 \
            touch(&timestamp);                                               \
        end_using_tree();                                                    \
        reset_mnemonics();                                                   \
        return timestamp != 0;                                               \
    } while (0)

gboolean rebuild_directories(GtkMenuItem * root)
{
    static GtkMenuItem *menu_item;
    static DECLARE_TIMESTAMP(timestamp);
    GtkMenuShell *menu;
    GNode *tree, *child;
    gchar *prefix, *old_name;
    tree_item *item;
    GSList *list = NULL;

    REBUILD_START(_("Directories"));

    /* Build the menu. */
    add_file_from_tree(NULL, NULL);

    item = tree->data;
    prefix = item->path;

    if (prefix[0] == '\0' || prefix[1] == '\0') {
        for (child = g_node_first_child(tree); child; child = child->next) {
            item = child->data;
            old_name = item->name;
            item->name = g_build_filename(prefix, item->name, NULL);

            make_menu_from_tree_rec(child, &list);

            g_free(item->name);
            item->name = old_name;
        }
    } else
        make_menu_from_tree_rec(tree, &list);

    REBUILD_END(rebuild_directories);
}

gboolean rebuild_images(GtkMenuItem * root)
{
    static GtkMenuItem *menu_item;
    static DECLARE_TIMESTAMP(timestamp);
    GSList *list = NULL;
    GtkMenuShell *menu;
    GNode *tree;

    REBUILD_START(_("Images"));

    /* Build the menu. */
    add_file_item(NULL, NULL);
    g_node_traverse(tree, G_PRE_ORDER, G_TRAVERSE_LEAFS, -1,
                    (GNodeTraverseFunc) add_file_item, &list);

    REBUILD_END(rebuild_images);
}

/* Rebuild both menus. */
gboolean rebuild_images_menus(void)
{
    return rebuild_directories(NULL) && rebuild_images(NULL);
}

void set_stop_rebuilding_menu(GtkMenuItem * item)
{
    cancel_menu_item = item;
    gtk_widget_set_sensitive(GTK_WIDGET(item), FALSE);
}
