/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <guichaz@yahoo.fr>
 */

/****************************
 * Which image to load next *
 ****************************/

#include "gliv.h"
#include "next_image.h"
#include "options.h"
#include "gliv_image.h"
#include "rendering.h"
#include "loading.h"
#include "files_list.h"
#include "textures.h"
#include "messages.h"
#include "main.h"
#include "menus.h"
#include "callbacks.h"

extern rt_struct *rt;
extern options_struct *options;
extern gliv_image *current_image;

/* We can have previous_image == next_image. */
static gliv_image *previous_image = NULL;
static gliv_image *next_image = NULL;

static GtkMenuItem *slide_show_menu;
static guint slide_show_timer = 0;

/*
 * Once the images list has been reordered we must take care
 * of the current image position and destroy the previous and
 * next images since they changed.
 */
void after_reorder(gint new_id)
{
    if (previous_image != NULL &&
        current_image->node->prev != previous_image->node) {

        destroy_image(previous_image);
        if (previous_image == next_image)
            next_image = NULL;

        previous_image = NULL;
    }

    if (next_image != NULL && current_image->node->next != next_image->node) {
        destroy_image(next_image);
        next_image = NULL;
    }

    current_image->number = new_id;
    fill_ident(current_image);

    refresh(REFRESH_STATUS);

    g_free(current_image->ident);
    current_image->ident = NULL;
}

/* Returns the previously shown image. */
static gliv_image *render_next_image(gliv_image * im)
{
    gliv_image *old = current_image;

    current_image = im;
    render();

    return old;
}

/*** Next/previous image ***/

static GList *next_node(GList * node, gint dir)
{
    if (node == NULL)
        return (dir == 1) ? get_list_head() : get_list_end();

    if (dir == 1)
        return node->next;

    /* dir == -1 */
    return node->prev;
}

static gliv_image *load_next_node(GList ** node, gint dir)
{
    gliv_image *im = NULL;
    GList *next;

    while ((next = next_node(*node, dir)) != NULL) {
        im = load_file(next->data);
        if (im == NULL)
            /* Delete bad images from the list. */
            remove_from_list(next);
        else {
            *node = next;
            break;
        }
    }

    return im;
}

static gliv_image *load_in_direction(gint dir)
{
    gliv_image *im = NULL;
    GList *node;
    gint id;

    if (get_list_length() == 2 && options->loop) {
        /*
         * When there are only two images, and we are looping,
         * next_image and previous_image are equal.
         */

        if (previous_image != NULL)
            return previous_image;

        if (next_image != NULL)
            return next_image;
    }

    if (current_image == NULL) {
        /* First time. */
        node = NULL;
        id = 0;
    } else {
        node = current_image->node;
        id = current_image->number + dir;
    }

    im = load_next_node(&node, dir);

    if (im == NULL && options->loop && get_list_length() > 1) {
        /* Loop at the end. */
        node = NULL;
        id = (dir == 1) ? 0 : (get_list_length() - 1);
        im = load_next_node(&node, dir);
    }

    if (im != NULL) {
        im->node = node;
        im->number = id;
    }

    return im;
}

/* Called when switching images. */
static void destroy_unused(gliv_image ** backup, gliv_image ** next)
{
    /*
     * If there are only two images,
     * we can have *backup == *next.
     */
    if (*backup != *next) {
        destroy_image(*backup);
        *backup = NULL;
    }

    if (options->one_image)
        destroy_image(current_image);
    else {
        prioritize_textures(current_image, FALSE);
        *backup = current_image;
    }
}

/* Return FALSE if there are no more images. */
gboolean load_direction(gint dir)
{
    gliv_image **backup;
    gliv_image **next;
    gliv_image *im;

    static volatile gint next_direction;

    /*
     * Before loading the next image in a prebuffer we let GTK process
     * remaining events and if there is a request for the next
     * image in the events pending then -> crash.
     * With this boolean we process loading request only if we know
     * all loadings are finished.
     * IOW: we are not reentrant.
     */
    static gboolean load_finished = TRUE;

    if (load_finished == FALSE) {
        /*
         * We come from process_events() called here, so we inform
         * that we want to load the next image.
         */
        next_direction = dir;
        return TRUE;
    }

    if (is_loading() || current_image == NULL)
        return TRUE;

    load_finished = FALSE;

    /* We process subsequent next image calls in the same loop. */
    while (dir * dir == 1) {

        if (dir < 0) {
            /* dir == -1 */
            backup = &next_image;
            next = &previous_image;
        } else {
            /* dir == 1 */
            backup = &previous_image;
            next = &next_image;
        }

        /*
         * next_direction can be changed by us when called
         * by process_events().
         */
        next_direction = 0;

        /* Be sure there is a next image. */
        if (*next == NULL)
            *next = load_in_direction(dir);

        im = *next;
        if (im == NULL) {
            /* There is no next image to load. */
            load_finished = TRUE;
            return FALSE;
        }

        destroy_unused(backup, next);
        *next = NULL;

        render_next_image(im);

        process_events();

        if (options->one_image == FALSE)
            *next = load_in_direction(dir);

        /* Usually next_direction == 0, and the while loop ends. */
        dir = next_direction;
    }

    load_finished = TRUE;
    return TRUE;
}

/*
 * After many images are opened we want to load
 * the next image but not the preloaded one.
 * Maybe the image after the one we will load does
 * not change.
 */
void open_next_image(gboolean keep_next)
{
    if (is_loading())
        return;

    if (keep_next) {
        if (previous_image != next_image)
            destroy_image(previous_image);
    } else
        unload_images();

    prioritize_textures(current_image, FALSE);
    previous_image = current_image;

    destroy_image(render_next_image(load_in_direction(1)));

    if (options->one_image == FALSE && next_image == NULL)
        next_image = load_in_direction(1);
}

/*** First/second image ***/

gliv_image *load_first_image(void)
{
    gliv_image *im;

    install_segv_handler();

    if (get_list_head() == NULL)
        return NULL;

    im = load_in_direction(1);
    if (im == NULL)
        DIALOG_MSG(_("No image found"));
    else
        render_next_image(im);

    return im;
}

void load_second_image(void)
{
    if (next_image == NULL && options->one_image == FALSE)
        next_image = load_in_direction(1);

    if (options->start_show)
        start_slide_show();
}

/*** Random loading ***/

void open_file(const gchar * filename, gboolean add_dir, gboolean shuffle_flag)
{
    gliv_image *im;
    gint nb_inserted;

    if (is_loading())
        return;

    /* Try to load the requested image. */
    im = load_file(filename);
    if (im == NULL)
        return;

    if (add_dir) {
        /* Add all files in the directory in the list. */
        gchar *dir = g_path_get_dirname(filename);

        nb_inserted = insert_after_current(&dir, 1, FALSE,
                                           shuffle_flag, !shuffle_flag, TRUE);
        g_free(dir);
        if (nb_inserted == 0)
            return;

        /* Put the requested image at current+1 position in the file list. */
        place_next(filename);
    } else {
        nb_inserted = insert_after_current((gchar **) & filename, 1, FALSE,
                                           shuffle_flag, !shuffle_flag, TRUE);
        if (nb_inserted == 0)
            return;
    }

    if (current_image == NULL) {
        im->node = get_list_head();
        im->number = 0;
    } else {
        im->node = current_image->node->next;
        im->number = current_image->number + 1;
    }

    open_next_image(nb_inserted == 1);
}

/*
 * Reloading.
 * Used for example, when an option is changed such as dithering or mipmap.
 */

static gliv_image *reload_image(gliv_image * im)
{
    gliv_image *new;

    if (im == NULL)
        return NULL;

    new = load_file(im->node->data);
    if (new == NULL)
        return im;

    new->node = im->node;
    new->number = im->number;

    destroy_image(im);

    return new;
}

void reload_all_images(void)
{
    if (is_loading())
        return;

    current_image = reload_image(current_image);
    prioritize_textures(current_image, TRUE);

    rt->filtering_enabled = FALSE;
    refresh(REFRESH_NOW);

    if (previous_image == next_image)
        previous_image = next_image = reload_image(previous_image);
    else {
        previous_image = reload_image(previous_image);
        next_image = reload_image(next_image);
    }
}

/*** Menu ***/

/* Check if we can optimize the loading asked by the images menu. */
static gboolean check_direction(const gchar * filename, gint dir)
{
    gliv_image **prev, **next;
    GList *next_node;

    if (dir == 1) {
        prev = &previous_image;
        next = &next_image;

        next_node = current_image->node->next;
        if (next_node != NULL)
            next_node = next_node->next;
    } else {
        /* dir == -1 */
        prev = &next_image;
        next = &previous_image;

        next_node = current_image->node->prev;
        if (next_node != NULL)
            next_node = next_node->prev;
    }

    if (*next == NULL)
        return FALSE;

    /* Check if the requested image is just following the current one. */
    if (filename == (*next)->node->data) {
        load_direction(dir);
        return TRUE;
    }

    /*
     * Check if the requested image is just before the previous one,
     * or just after the next one.
     */
    if (next_node != NULL && filename == next_node->data) {
        gint number;
        gliv_image *new;

        destroy_image(*prev);
        *prev = *next;
        *next = NULL;

        number = current_image->number + 2 * dir;

        new = load_file(filename);
        if (new == NULL) {
            remove_from_list(next_node);
            return TRUE;
        }

        new->node = next_node;
        new->number = number;

        destroy_image(render_next_image(new));
        return TRUE;
    }

    return FALSE;
}

void menu_load(const gchar * filename)
{
    GList *node;
    gint number = 0;
    gliv_image *im;
    static gboolean loading = FALSE;

    if (loading) {
        /* We are not reentrant. */
        return;
    }

    loading = TRUE;

    if (is_loading() || filename == current_image->node->data ||
        check_direction(filename, -1) || check_direction(filename, 1)) {
        loading = FALSE;
        return;
    }

    for (node = get_list_head(); node != NULL; node = node->next) {
        if (filename == node->data)
            break;

        number++;
    }

    if (node == NULL) {
        /* The image is not in the list anymore. */
        loading = FALSE;
        return;
    }

    im = load_file(filename);
    if (im == NULL) {
        remove_from_list(node);
        loading = FALSE;
        return;
    }

    unload_images();

    im->node = node;
    im->number = number;

    destroy_image(render_next_image(im));
    loading = FALSE;
}

/* Called when destroying a file. */
void unload(GList * node)
{
    if (previous_image != NULL && previous_image->node == node) {
        destroy_image(previous_image);
        if (previous_image == next_image)
            next_image = NULL;

        previous_image = NULL;

    } else if (next_image != NULL && next_image->node == node) {
        destroy_image(next_image);
        next_image = NULL;
    }
}

/* Called when we want only one image in memory. */
void unload_images(void)
{
    destroy_image(previous_image);

    if (previous_image != next_image)
        destroy_image(next_image);

    previous_image = NULL;
    next_image = NULL;
}

/* Called when we insert a GLiv collection. */
void destroy_next_image(void)
{
    destroy_image(next_image);
    next_image = NULL;
}

/*** Slide show ***/

void set_slide_show_menu(GtkMenuItem * item)
{
    slide_show_menu = item;
}

static void stop_slide_show(void)
{
    if (slide_show_started()) {
        g_source_remove(slide_show_timer);
        slide_show_timer = 0;
    }

    set_menu_label(slide_show_menu, _("Start the slide show"), TRUE);
}

static gboolean slide_show_next(void)
{
    if (load_direction(1) == FALSE) {
        /* No more images. */
        stop_slide_show();
        return FALSE;
    }

    return TRUE;
}

void start_slide_show(void)
{
    stop_slide_show();

    if (options->duration < 0 || current_image == NULL)
        return;

    slide_show_timer = g_timeout_add(options->duration * 1000,
                                     (GSourceFunc) slide_show_next, NULL);

    set_menu_label(slide_show_menu, _("Stop the slide show"), TRUE);
}

gboolean slide_show_started(void)
{
    return slide_show_timer != 0;
}

void toggle_slide_show(void)
{
    if (slide_show_started())
        stop_slide_show();
    else
        start_slide_show();
}
